from datetime import datetime
from string import printable, whitespace

from bleak.backends.scanner import AdvertisementData
from bleak.uuids import uuidstr_to_str
from bluetooth_numbers.companies import company
from rich._palettes import EIGHT_BIT_PALETTE
from rich.style import Style
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from humble_explorer.utils import hash8

__author__ = "Koen Vervloesem"
__copyright__ = "Koen Vervloesem"
__license__ = "MIT"

printable_chars = printable.replace(whitespace, " ")


class Now:
    """Rich renderable that shows the current time. All times within the same second
    are rendered in the same color."""

    def __init__(self, style: Style = None):
        now = datetime.now()
        self.time = now.strftime("%H:%M:%S.%f")
        if style:
            self.style = style
        else:
            self.style = Style(
                color=EIGHT_BIT_PALETTE[hash8(now.strftime("%H:%M:%S"))].hex
            )

    def __rich__(self) -> Text:
        return Text(self.time, style=self.style)


class DeviceAddress:
    """Rich renderable that shows a Bluetooth device address. Every address is rendered
    in its own color."""

    def __init__(self, address: str):
        self.address = address
        self.style = Style(color=EIGHT_BIT_PALETTE[hash8(self.address)].hex)

    def __rich__(self) -> Text:
        return Text(self.address, style=self.style)


class RSSI:
    """Rich renderable that shows RSSI of a device."""

    def __init__(self, rssi: int):
        self.rssi = rssi

    def __rich__(self) -> Text:
        return Text.assemble((str(self.rssi), "green bold"), " dBm")


class UUID:
    """Rich renderable that shows a UUID with description and colors."""

    def __init__(self, uuid128: str):
        self.uuid128 = uuid128

    def __rich__(self) -> Text:
        # Colorize the 16-bit UUID part in a standardized 128-bit UUID.
        if self.uuid128.startswith("0000") and self.uuid128.endswith(
            "-0000-1000-8000-00805f9b34fb"
        ):
            colored_uuid = Text.assemble(
                "0000",
                (self.uuid128[4:8], "green bold"),
                "-0000-1000-8000-00805f9b34fb",
            )
        else:
            colored_uuid = self.uuid128

        return Text.assemble(colored_uuid, f" ({uuidstr_to_str(self.uuid128)})")


class CompanyID:
    """Rich renderable that shows company ID and name."""

    def __init__(self, cic: int):
        self.cic = cic

    def __rich__(self) -> Text:
        try:
            manufacturer_name = company[self.cic]
        except KeyError:
            manufacturer_name = "Unknown"

        return Text.assemble(
            f"0x{self.cic:04x} (", (manufacturer_name, "green bold"), ")"
        )


class HexData:
    """Rich renderable that shows hex data."""

    def __init__(self, data: bytes):
        self.data = data

    def __rich__(self) -> Text:
        return Text(f"{self.data.hex(sep= ' ')}", style="cyan bold")


class HexString:
    """Rich renderable that shows hex data as a string with non-printable characters
    replaced by a dot."""

    def __init__(self, data: bytes):
        self.data = data

    def __rich__(self) -> str:
        result = []
        for byte in self.data:
            char = chr(byte)
            if char in printable_chars:
                result.append(f" {char}")
            else:
                result.append(" .")

        return " ".join(result)


class RichAdvertisement:
    """Rich renderable that shows advertisement data."""

    def __init__(self, data: AdvertisementData):
        self.data = data

    def __rich__(self) -> Table:
        table = Table(show_header=False, show_edge=False, padding=0)

        # Show local name
        if self.data.local_name:
            table.add_row(
                Text.assemble("local name: ", (self.data.local_name, "green bold"))
            )

        # Show RSSI
        if self.data.rssi:
            table.add_row(Text.assemble("RSSI: ", RSSI(self.data.rssi).__rich__()))

        # Show TX Power
        if self.data.tx_power:
            table.add_row(
                Text.assemble("TX Power: ", RSSI(self.data.tx_power).__rich__())
            )

        # Show manufacturer data
        if self.data.manufacturer_data:
            tree = Tree("manufacturer data:")
            for cic, value in self.data.manufacturer_data.items():
                company = Tree(
                    Text.assemble(CompanyID(cic).__rich__(), f" → {len(value)} bytes")
                )
                company.add(Text.assemble("hex  → ", HexData(value).__rich__()))
                company.add(Text.assemble("text → ", HexString(value).__rich__()))
                tree.add(company)
            table.add_row(tree)

        # Show service data
        if self.data.service_data:
            tree = Tree("service data:")
            for uuid, value in self.data.service_data.items():
                svc_uuid = Tree(
                    Text.assemble(UUID(uuid).__rich__(), f" → {len(value)} bytes")
                )
                svc_uuid.add(Text.assemble("hex  → ", HexData(value).__rich__()))
                svc_uuid.add(Text.assemble("text → ", HexString(value).__rich__()))
                tree.add(svc_uuid)
            table.add_row(tree)

        # Show service UUIDs with their description
        if self.data.service_uuids:
            tree = Tree("service UUIDs:")
            for uuid in sorted(self.data.service_uuids):
                tree.add(UUID(uuid))
            table.add_row(tree)

        return table

    def height(self) -> int:
        height = 0
        if self.data.local_name:
            height += 1
        if self.data.rssi:
            height += 1
        if self.data.tx_power:
            height += 1
        if self.data.manufacturer_data:
            height = height + 1 + 3 * len(self.data.manufacturer_data)
        if self.data.service_data:
            height = height + 1 + 3 * len(self.data.service_data)
        if self.data.service_uuids:
            height = height + 1 + len(self.data.service_uuids)
        return height



from rich.text import Text

from humble_explorer.renderables import RSSI, Now

__author__ = "Koen Vervloesem"
__copyright__ = "Koen Vervloesem"
__license__ = "MIT"


def test_now():
    """Test Now class."""
    now1 = Now()
    now2 = Now()
    # make now1 and now2 times within the same second
    while now1.time.split(".")[0] != now2.time.split(".")[0]:
        now1 = Now()

    assert now1.style == now2.style


def test_rssi():
    """Test RSSI class."""
    rssi = RSSI(-70)
    assert rssi.__rich__() == Text.assemble(("-70", "green bold"), " dBm")



from argparse import Namespace
from platform import system

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from rich.style import Style
from textual.app import App, ComposeResult
from textual.widgets import DataTable, Footer, Header

if system() == "Linux":
    from bleak.assigned_numbers import AdvertisementDataType
    from bleak.backends.bluezdbus.advertisement_monitor import OrPattern
    from bleak.backends.bluezdbus.scanner import BlueZScannerArgs

from humble_explorer.renderables import DeviceAddress, Now, RichAdvertisement

__author__ = "Koen Vervloesem"
__copyright__ = "Koen Vervloesem"
__license__ = "MIT"

_PAUSE_STYLE = Style(color="red", bgcolor="grey50")


class BLEScannerApp(App[None]):
    """A Textual app to scan for Bluetooth Low Energy advertisements."""

    TITLE = "HumBLE Explorer"
    BINDINGS = [
        ("q", "quit", "Quit"),
        ("s", "toggle_scan", "Toggle scan"),
    ]

    def __init__(self, cli_args: Namespace):
        """Initialize BLE scanner."""

        # Configure scanning mode
        scanner_kwargs = {"scanning_mode": cli_args.scanning_mode}

        # Passive scanning with BlueZ needs at least one or_pattern.
        # The following matches all devices.
        if system() == "Linux" and cli_args.scanning_mode == "passive":
            scanner_kwargs["bluez"] = BlueZScannerArgs(
                or_patterns=[
                    OrPattern(0, AdvertisementDataType.FLAGS, b"\x06"),
                    OrPattern(0, AdvertisementDataType.FLAGS, b"\x1a"),
                ]
            )

        # Configure Bluetooth adapter
        scanner_kwargs["adapter"] = cli_args.adapter

        # Set up BleakScanner object
        scanner_kwargs["detection_callback"] = self.on_advertisement
        self.scanner = BleakScanner(**scanner_kwargs)
        self.scanning = False

        super().__init__()

    async def action_toggle_scan(self) -> None:
        """Start or stop BLE scanning."""
        if self.scanning:
            await self.stop_scan()
        else:
            await self.start_scan()

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Footer()
        yield DataTable(zebra_stripes=True)

    async def on_advertisement(
        self, device: BLEDevice, advertisement_data: AdvertisementData
    ) -> None:
        """Show advertisement data on detection of a BLE advertisement."""
        self.log(advertisement_data.local_name, device.address, advertisement_data)
        rich_advertisement = RichAdvertisement(advertisement_data)
        table = self.query_one(DataTable)
        table.add_row(
            Now(),
            DeviceAddress(device.address),
            rich_advertisement,
            height=rich_advertisement.height(),
        )
        table.scroll_end(animate=False)

    async def on_mount(self) -> None:
        """Initialize interface and start BLE scan."""
        table = self.query_one(DataTable)
        table.add_columns("Time", "Address", "Advertisement")
        await self.start_scan()

    async def start_scan(self) -> None:
        """Start BLE scan."""
        self.scanning = True
        await self.scanner.start()

    async def stop_scan(self) -> None:
        """Stop BLE scan."""
        self.scanning = False
        await self.scanner.stop()
        table = self.query_one(DataTable)
        table.add_row(Now(style=_PAUSE_STYLE))
        table.scroll_end(animate=False)
